from __future__ import annotations

from dataclasses import dataclass, field
from typing import Dict, Union

import pytest

from draccus.utils import DecodingError

from .testutils import *


def test_union_type():
    @dataclass
    class Foo(TestSetup):
        x: Union[float, str] = 0

    Foo.get_help_text()

    foo = Foo.setup("--x 1.2")
    assert foo.x == 1.2

    foo = Foo.setup("--x bob")
    assert foo.x == "bob"


@pytest.mark.skipif(sys.version_info < (3, 10), reason="requires python3.10 or higher")
def test_union_types_39():
    @dataclass
    class Foo(TestSetup):
        x: float | str = 0

    foo = Foo.setup("--x 1.2")
    assert foo.x == 1.2

    foo = Foo.setup("--x bob")
    assert foo.x == "bob"


@pytest.mark.skipif(sys.version_info < (3, 10), reason="requires python3.10 or higher")
def test_union_types_39_optional():
    @dataclass
    class Foo(TestSetup):
        x: Optional[float | str] = 0

    foo = Foo.setup("--x 1.2")
    assert foo.x == 1.2

    foo = Foo.setup("--x bob")
    assert foo.x == "bob"

    foo = Foo.setup("--x null")
    assert foo.x is None


@pytest.mark.skipif(sys.version_info < (3, 10), reason="requires python3.10 or higher")
def test_union_types_39_optional_nested():
    @dataclass
    class Foo(TestSetup):
        x: Union[float, int | str] = 0

    foo = Foo.setup("--x 1.2")
    assert foo.x == 1.2

    foo = Foo.setup("--x bob")
    assert foo.x == "bob"


def test_union_error_message_atomics():
    @dataclass
    class Foo(TestSetup):
        x: Union[float, bool] = 0

    with pytest.raises(DecodingError) as e:
        Foo.setup("--x 1.2.3")

    assert """`x`: Could not decode the value into any of the given types:
    float: Couldn't parse '1.2.3' into a float
    bool: Couldn't parse '1.2.3' into a bool""" in str(
        e.value
    )


def test_union_error_message_nested():
    @dataclass
    class Foo(TestSetup):
        x: Union[float, Union[int, bool]] = 0

    with pytest.raises(DecodingError) as e:
        Foo.setup("--x 1.2.3")

    assert """`x`: Could not decode the value into any of the given types:
    float: Couldn't parse '1.2.3' into a float
    int: Couldn't parse '1.2.3' into an int
    bool: Couldn't parse '1.2.3' into a bool""" in str(
        e.value
    )


@dataclass(frozen=True)
class Baz_u:
    z: int


@dataclass
class Foo_u(TestSetup):
    x: Union[bool, Baz_u]


def test_decode_union_with_dataclass_and_atomic():
    foo = Foo_u.setup("--x false")
    assert foo.x is False

    foo = Foo_u.setup("--x.z 1")
    assert foo.x == Baz_u(z=1)

    try:
        foo = Foo_u.setup("--x.z 1.2")
        raise AssertionError()
    except DecodingError:
        pass


@dataclass(frozen=True)
class Baz_e:
    z: int
    y: str


@dataclass(frozen=True)
class Bar_e:
    z: bool


@dataclass
class Foo_e(TestSetup):
    x: Union[Baz_e, Bar_e] = Bar_e(False)


def test_union_error_message_dataclasses():
    with pytest.raises(DecodingError) as e:
        Foo_e.setup("--x.z 1.2.3")

    assert """`x`: Could not decode the value into any of the given types:
    Baz_e: `z`: Couldn't parse '1.2.3' into an int
    Bar_e: `z`: Couldn't parse '1.2.3' into a bool""".strip() in str(
        e.value
    )

    with pytest.raises(DecodingError) as e:
        Foo_e.setup("--x.y foo")

    assert """`x`: Could not decode the value into any of the given types:
    Baz_e: Missing required field(s) `z` for Baz_e
    Bar_e: The fields `y` are not valid for Bar_e""".strip() in str(
        e.value
    )


@dataclass
class Bar:
    y: int


@dataclass
class Foo(TestSetup):
    x: Optional[Union[Bar, Dict[str, Bar]]] = field(default=None)


def test_union_argparse_dict():
    foo = Foo.setup('--x \'{"a": {"y": 1}, "b": {"y": 2}}\'')
    assert foo.x == {"a": Bar(y=1), "b": Bar(y=2)}



# test_optional_choice_type.py

from dataclasses import dataclass
from typing import Optional

import pytest

import draccus
from draccus.choice_types import ChoiceRegistry
from draccus.utils import DecodingError
from tests.testutils import TestSetup


@dataclass
class Person(ChoiceRegistry):
    name: str


@dataclass
class Adult(Person):
    age: int


@dataclass
class Child(Person):
    favorite_toy: str


Person.register_subclass("adult", Adult)
Person.register_subclass("child", Child)


@dataclass
class Profile(TestSetup):
    person: Optional[Person] = None


def test_optional_choice_empty():
    profile = Profile.setup("")
    assert profile.person is None


def test_optional_choice_adult():
    profile = Profile.setup("--person.type adult --person.name Bob --person.age 30")
    assert profile.person == Adult(name="Bob", age=30)


def test_optional_choice_child():
    profile = Profile.setup("--person.type child --person.name Alice --person.favorite_toy truck")
    assert profile.person == Child(name="Alice", favorite_toy="truck")


def test_invalid_choice():
    with pytest.raises(SystemExit) as excinfo:
        Profile.setup("--person.type invalid_type --person.name Jill")
    assert excinfo.type is SystemExit
    assert excinfo.value.code == 2


def test_invalid_fields_adult():
    with pytest.raises(DecodingError):
        Profile.setup("--person.type adult --person.name Bob --person.age 30 --person.favorite_toy truck")


def test_encode_optional_none():
    profile = Profile()
    assert draccus.encode(profile) == {"person": None}


def test_encode_optional_child():
    profile = Profile(person=Child(name="Kevin", favorite_toy="ball"))
    encoded = draccus.encode(profile)
    assert encoded == {
        "person": {
            "type": "child",
            "name": "Kevin",
            "favorite_toy": "ball",
        }
    }


def test_encode_optional_adult():
    profile = Profile(person=Adult(name="Bob", age=42))
    encoded = draccus.encode(profile)
    assert encoded == {
        "person": {
            "type": "adult",
            "name": "Bob",
            "age": 42,
        }
    }



from typing import Type

from draccus import utils
from draccus.options import Options, config_type
from draccus.parsers.decoding import decode
from draccus.parsers.encoding import encode
from draccus.utils import Dataclass


def parse_string(s):
    parser = Options.get_config_type().value
    return parser.parse_string(s)


def load_config(stream, *, file=None):
    if file is not None:
        fpath = str(file)
        if fpath.endswith(".toml"):
            with config_type("toml"):
                return load_config(stream)
        elif fpath.endswith(".json"):
            with config_type("json"):
                return load_config(stream)
        elif fpath.endswith(".yaml") or fpath.endswith(".yml"):
            with config_type("yaml"):
                return load_config(stream)
    parser = Options.get_config_type().value
    return parser.load_config(stream)


def save_config(d, stream=None, **kwargs):
    parser = Options.get_config_type().value
    return parser.save_config(d, stream, **kwargs)


def load(t: Type[Dataclass], stream):
    dictionary = load_config(stream)
    return decode(t, dictionary)


def dump(config: Dataclass, stream=None, omit_defaults: bool = False, **kwargs):
    """
    Dump the config object to yaml.
    optionally omit any value that still has a default value
    """
    config_dict = encode(config)
    if omit_defaults:
        defaults_dict = encode(utils.get_defaults_dict(config))
        config_dict = utils.remove_matching(config_dict, defaults_dict)
    return save_config(config_dict, stream, **kwargs)



# tests that draccus can correctly parse and use generic parameters
from dataclasses import dataclass
from typing import Generic, List, TypeVar

import draccus

T = TypeVar("T")


@dataclass
class ScheduleStep(Generic[T]):
    until: int
    value: T


@dataclass
class Schedule(Generic[T]):
    phases: List[ScheduleStep[T]]


def test_generic_params():
    decoded = draccus.decode(Schedule[int], {"phases": [{"until": 10, "value": 1}]})

    assert decoded == Schedule(phases=[ScheduleStep(until=10, value=1)])

    encoded = draccus.encode(Schedule[int](phases=[ScheduleStep(until=10, value=1)]))

    assert encoded == {"phases": [{"until": 10, "value": 1}]}

    list_decoded = draccus.decode(List[ScheduleStep[int]], [{"until": 10, "value": 1}])

    assert list_decoded == [ScheduleStep(until=10, value=1)]



from dataclasses import dataclass, field
from typing import *

import draccus.utils as utils

from .testutils import *


@dataclass
class SomeDataclass:
    x: float = 123


@parametrize(
    "t",
    [
        Tuple[int, ...],
        Tuple[str],
        Tuple,
        tuple,
    ],
)
def test_is_tuple(t: Type):
    assert utils.is_tuple(t)
    assert not utils.is_list(t)


@parametrize(
    "t",
    [
        List[int],
        List[str],
        List,
        list,
        List[SomeDataclass],
    ],
)
def test_is_list(t: Type):
    assert utils.is_list(t)
    assert not utils.is_tuple(t)


@parametrize(
    "t",
    [
        List[SomeDataclass],
        Tuple[SomeDataclass],
    ],
)
def test_is_list_of_dataclasses(t: Type):
    assert utils.is_tuple_or_list_of_dataclasses(t)


@dataclass
class A:
    a: str = "bob"


import enum


class Color(enum.Enum):
    RED = "RED"
    ORANGE = "ORANGE"
    BLUE = "BLUE"


class Temperature(enum.IntEnum):
    HOT = 1
    WARM = 0
    COLD = -1
    MONTREAL = -35


@parametrize(
    "t",
    [
        Color,
        Temperature,
    ],
)
def test_is_enum(t: Type):
    assert utils.is_enum(t)


def test_list_field():
    @dataclass
    class A:
        a: List[str] = field(default_factory=["bob", "john", "bart"].copy)

    a1 = A()
    a2 = A()
    assert id(a1.a) != id(a2.a)


def test_set_field():
    @dataclass
    class A:
        a: Set[str] = field(default_factory=lambda: {"bob", "john", "bart"})

    a1 = A()
    a2 = A()
    assert id(a1.a) != id(a2.a)


def test_dict_field():
    default = {"bob": 0, "john": 1, "bart": 2}

    @dataclass
    class A:
        a: Dict[str, int] = field(default_factory=default.copy)

    a1 = A()
    print(a1.a)
    assert a1.a == default
    a2 = A()
    assert id(a1.a) != id(a2.a)


def test_dict_field_with_keyword_args():
    default = {"bob": 0, "john": 1, "bart": 2}

    @dataclass
    class A(TestSetup):
        a: Dict[str, int] = field(default_factory=dict(bob=0, john=1, bart=2).copy)

    a1 = A()
    a2 = A()
    assert a1.a == a2.a == default
    assert id(a1.a) != id(a2.a)


def test_dict_field_without_args():
    default = {}

    @dataclass
    class A(TestSetup):
        a: Dict[str, int] = field(default_factory=dict)

    a1 = A()
    a2 = A()
    assert a1.a == a2.a == default
    assert id(a1.a) != id(a2.a)



import contextlib
from typing import Union

from draccus.parsers.config_parsers import JSONParser, ParserEnum, TOMLParser, YAMLParser


class ConfigType(ParserEnum):
    YAML = YAMLParser
    JSON = JSONParser
    TOML = TOMLParser


class Options:
    _config_type: ConfigType = ConfigType.YAML

    @staticmethod
    def set_config_type(new_type: Union[ConfigType, str]):
        if isinstance(new_type, str):
            new_type = ConfigType[new_type.upper()]
        Options._config_type = new_type

    @staticmethod
    def get_config_type() -> ConfigType:
        return Options._config_type


@contextlib.contextmanager
def config_type(new_type: Union[ConfigType, str]):
    current_type = Options.get_config_type()
    try:
        Options.set_config_type(new_type)
        yield
    finally:
        Options.set_config_type(current_type)



from dataclasses import dataclass, field
from enum import Enum, auto
from typing import List

from draccus import ParsingError

from .testutils import TestSetup, raises


class Color(Enum):
    blue: str = auto()  # type: ignore
    red: str = auto()  # type: ignore
    green: str = auto()  # type: ignore
    orange: str = auto()  # type: ignore


class BigColor(Enum):
    BLUE: str = "blue"
    RED: str = "red"
    GREEN: str = "green"
    TRICKY: str = "orange"


def test_passing_enum_to_choice():
    @dataclass
    class Something(TestSetup):
        favorite_color: Color = field(default=Color.green)
        colors: List[Color] = field(default_factory=[Color.green].copy)

    s = Something.setup("")
    assert s.favorite_color == Color.green
    assert s.colors == [Color.green]

    s = Something.setup("--colors [blue,red]")
    assert s.colors == [Color.blue, Color.red]


def test_passing_enum_to_choice_from_value():
    @dataclass
    class Something(TestSetup):
        favorite_color: BigColor = field(default=BigColor.GREEN)
        colors: List[BigColor] = field(default_factory=[BigColor.GREEN].copy)

    s = Something.setup("")
    assert s.favorite_color == BigColor.GREEN
    assert s.colors == [BigColor.GREEN]

    s = Something.setup("--colors '[blue, red]'")
    assert s.colors == [BigColor.BLUE, BigColor.RED]

    # try tricky
    s = Something.setup("--colors '[blue,orange]'")
    assert s.colors == [BigColor.BLUE, BigColor.TRICKY]


#
#
def test_passing_enum_to_choice_no_default_makes_required_arg():
    @dataclass
    class Something(TestSetup):
        favorite_color: Color = field()

    with raises(ParsingError):
        s = Something.setup("")

    s = Something.setup("--favorite_color blue")
    assert s.favorite_color == Color.blue


def test_passing_enum_to_choice_is_same_as_enum_attr():
    @dataclass
    class Something1(TestSetup):
        favorite_color: Color = Color.orange

    @dataclass
    class Something2(TestSetup):
        favorite_color: Color = field(default=Color.orange)

    s1 = Something1.setup("--favorite_color green")
    s2 = Something2.setup("--favorite_color green")
    assert s1.favorite_color == s2.favorite_color

    s = Something1.setup("--favorite_color blue")
    assert s.favorite_color == Color.blue
    s = Something2.setup("--favorite_color blue")
    assert s.favorite_color == Color.blue



from abc import get_cache_token
from dataclasses import dataclass
from functools import _find_impl, update_wrapper  # type: ignore
from typing import Callable, Optional

from draccus.utils import canonicalize_union


@dataclass
class RegistryFunc:
    # The function saved in the registry
    func: Callable
    # Whether the function should be registered for subclasses as well
    include_subclasses: bool


def withregistry(base_func):
    import types
    import weakref

    registry = {}
    dispatch_cache = weakref.WeakKeyDictionary()
    cache_token = None

    def dispatch(cls) -> Optional[RegistryFunc]:
        nonlocal cache_token
        # Python 3.10: you can't have weak refs to x|y union types, so we need to canonicalize
        cls = canonicalize_union(cls)
        if cache_token is not None:
            current_token = get_cache_token()
            if cache_token != current_token:
                dispatch_cache.clear()
                cache_token = current_token
        if cls in dispatch_cache:
            impl = dispatch_cache[cls]
        else:
            if cls in registry:
                impl = registry[cls]
            else:
                try:
                    impl = _find_impl(cls, registry)
                    if not impl.include_subclasses:
                        # Do not allow implicit inherited implementation without type
                        impl = None
                except Exception:
                    impl = None
            dispatch_cache[cls] = impl

        return impl

    def register(cls, func=None, include_subclasses=False):
        nonlocal cache_token
        if func is None:
            if isinstance(cls, type):
                return lambda f: register(cls, func=f, include_subclasses=include_subclasses)
            ann = getattr(cls, "__annotations__", {})
            if not ann:
                raise TypeError(
                    f"Invalid first argument to `register()`: {cls!r}. "
                    "Use either `@register(some_class)` or plain `@register` "
                    "on an annotated function."
                )
            func = cls

            # only import typing if annotation parsing is necessary
            from typing import get_type_hints

            argname, cls = next(iter(get_type_hints(func).items()))
            assert isinstance(cls, type), f"Invalid annotation for {argname!r}. {cls!r} is not a class."
        registry[cls] = RegistryFunc(func, include_subclasses)
        if cache_token is None and hasattr(cls, "__abstractmethods__"):
            cache_token = get_cache_token()
        dispatch_cache.clear()
        return func

    def wrapper(*args, **kw):
        # Unlike singledispatch we do not directly override the base call
        return base_func(*args, **kw)

    wrapper.register = register
    wrapper.dispatch = dispatch
    wrapper.registry = types.MappingProxyType(registry)
    wrapper._clear_cache = dispatch_cache.clear
    update_wrapper(wrapper, base_func)
    return wrapper



"""Simple, Elegant Argument parsing.
@author: Fabrice Normandin
"""
import argparse
import dataclasses
import inspect
import os
import sys
import warnings
from argparse import HelpFormatter, Namespace
from collections import defaultdict
from functools import wraps
from gettext import gettext
from logging import getLogger
from pathlib import Path
from typing import Dict, Generic, Optional, Sequence, Text, Type, TypeVar, Union

import mergedeep

from draccus import cfgparsing, utils
from draccus.help_formatter import SimpleHelpFormatter
from draccus.parsers import decoding
from draccus.utils import Dataclass, DraccusException
from draccus.wrappers import DataclassWrapper
from draccus.wrappers.docstring import HelpOrder
from draccus.wrappers.suppressing_argparse import SuppressingArgumentParser

logger = getLogger(__name__)

T = TypeVar("T")


class ArgumentParser(Generic[T]):
    def __init__(
        self,
        config_class: Type[T],
        config_path: Optional[Union[Path, str]] = None,
        formatter_class: Type[HelpFormatter] = SimpleHelpFormatter,
        preferred_help: str = HelpOrder.inline,
        *args,
        **kwargs,
    ):
        """Creates an ArgumentParser instance."""
        kwargs = kwargs.copy()
        kwargs["formatter_class"] = formatter_class
        if "exit_on_error" in kwargs:
            # only available in python 3.9+, remove arg if not supported
            if sys.version_info < (3, 9):
                # behavior was functionally True before 3.9
                if not kwargs["exit_on_error"]:
                    warnings.warn(
                        "ArgumentParser exit_on_error is only available in python 3.9+, removing argument", stacklevel=2
                    )
                del kwargs["exit_on_error"]

        self.parser = SuppressingArgumentParser(*args, **kwargs)

        # constructor arguments for the dataclass instances.
        # (a Dict[dest, [attribute, value]])
        self.constructor_arguments: Dict[str, Dict] = defaultdict(dict)

        self.config_path = config_path
        self.config_class = config_class
        self.preferred_help = preferred_help
        self._assert_preferred_help()

        self._assert_no_conflicts()
        self.parser.add_argument(
            f"--{utils.CONFIG_ARG}",
            type=str,
            help="Path for a config file to parse with draccus",
        )
        self._set_dataclass(config_class)  # type: ignore

    def _set_dataclass(
        self,
        dataclass: Union[Type[Dataclass], Dataclass],
        default: Optional[Union[Dataclass, Dict]] = None,
        dataclass_wrapper_class: Type[DataclassWrapper] = DataclassWrapper,
    ):
        """Adds command-line arguments for the fields of `dataclass`."""
        if not isinstance(dataclass, type):
            default = dataclass if default is None else default
            dataclass = type(dataclass)

        new_wrapper = dataclass_wrapper_class(dataclass, default=default, preferred_help=self.preferred_help)
        new_wrapper.register_actions(parser=self.parser)

    def _assert_preferred_help(self):
        """Checks that `self.prefer_help` is valid."""
        if self.preferred_help not in {"inline", "above", "below"}:
            raise DraccusException(
                f"Value `prefer_help = {self.preferred_help}` not supported; must be one of < inline | above | below >"
            )

    def _assert_no_conflicts(self):
        """Checks for a field name that conflicts with utils.CONFIG_ARG"""
        if utils.CONFIG_ARG in [field.name for field in dataclasses.fields(self.config_class)]:
            raise DraccusException(f"{utils.CONFIG_ARG} is a reserved word for draccus")

    def parse_args(self, args=None, namespace=None) -> T:
        args, _ = self.parse_known_args(args, namespace, is_parse_args=True)
        return args

    def parse_known_args(
        self,
        args: Optional[Sequence[Text]] = None,
        namespace: Optional[Namespace] = None,
        *,
        is_parse_args: bool = False,
    ):
        # NOTE: since the usual ArgumentParser.parse_args() calls
        # parse_known_args, we therefore just need to overload the
        # parse_known_args method to support both.
        if args is None:
            # args default to the system args
            args = sys.argv[1:]
        else:
            # make sure that args are mutable
            args = list(args)

        if "--help" not in args and "-h" not in args:
            for action in self.parser._actions:
                # TODO(dlwh): this is so gross
                # TODO: Find a better way to do that?
                action.default = argparse.SUPPRESS  # To avoid setting of defaults in actual run
                action.type = str  # In practice, we want all processing to happen with yaml
        parsed_args, unparsed_args = self.parser.parse_known_args(args, namespace)
        if is_parse_args and unparsed_args:
            msg = gettext("unrecognized arguments: %s") % " ".join(unparsed_args)
            if getattr(self.parser, "exit_on_error", True):
                self.parser.error(msg)
            else:
                raise DraccusException(msg)

        parsed_t = self._postprocessing(parsed_args)
        return parsed_t, unparsed_args

    def print_help(self, file=None):
        return super().print_help(file)

    def _postprocessing(self, parsed_args: Namespace) -> T:
        logger.debug("\nPOST PROCESSING\n")
        logger.debug(f"(raw) parsed args: {parsed_args}")

        parsed_arg_values = vars(parsed_args)

        for key in parsed_arg_values:
            parsed_value = cfgparsing.parse_string(parsed_arg_values[key])
            if isinstance(parsed_value, str) and parsed_value.startswith("include"):
                try:
                    with open(parsed_value[len("include ") :], "r") as f:
                        parsed_arg_values[key] = cfgparsing.load_config(f)
                except FileNotFoundError as e:
                    raise FileNotFoundError(
                        f"{e}. Include is a reserved cli keyword. "
                        "If your argument uses include, "
                        "Please refer to https://github.com/dlwh/draccus/issues/17"
                    ) from e
            else:
                parsed_arg_values[key] = parsed_value

        config_path = self.config_path  # Could be NONE

        if utils.CONFIG_ARG in parsed_arg_values:
            new_config_path = parsed_arg_values[utils.CONFIG_ARG]
            if config_path is not None:
                warnings.warn(UserWarning(f"Overriding default {config_path} with {new_config_path}"), stacklevel=2)
            config_path = new_config_path
            del parsed_arg_values[utils.CONFIG_ARG]

        if config_path is not None:
            with open(config_path, "r") as f:
                file_args = cfgparsing.load_config(f, file=config_path)
        else:
            file_args = {}

        deflat_d = utils.deflatten(parsed_arg_values, sep=".")
        deflat_d = mergedeep.merge(file_args, deflat_d)
        cfg = decoding.decode(self.config_class, deflat_d)

        return cfg


def parse(
    config_class: Type[T],
    config_path: Optional[Union[Path, str]] = None,
    args: Optional[Sequence[str]] = None,
    prog: Optional[str] = None,
    exit_on_error: bool = True,
    preferred_help: str = HelpOrder.inline,
) -> T:
    """
    Parses the command line arguments and returns an instance of the config class.

    Args:
        config_class: The config class to parse.
        config_path: The path to the config file to parse.
        args: The command line arguments to parse. If None, defaults to sys.argv[1:].
        prog: The name of the program (for the help message).
        exit_on_error: Whether to exit if an error occurs.
        preferred_help: Preferred location to parse help text for fields (< "inline" | "above" | "below" >)
    """
    parser = ArgumentParser(
        config_class=config_class,
        config_path=config_path,
        exit_on_error=exit_on_error,
        prog=prog,
        preferred_help=preferred_help,
    )
    return parser.parse_args(args)


def wrap(config_path: Optional[os.PathLike] = None, preferred_help: str = HelpOrder.inline):
    def wrapper_outer(fn):
        @wraps(fn)
        def wrapper_inner(*args, **kwargs):
            argspec = inspect.getfullargspec(fn)
            argtype = argspec.annotations[argspec.args[0]]
            if len(args) > 0 and type(args[0]) is argtype:
                cfg = args[0]
                args = args[1:]
            else:
                cfg = parse(config_class=argtype, config_path=config_path, preferred_help=preferred_help)
            response = fn(cfg, *args, **kwargs)
            return response

        return wrapper_inner

    return wrapper_outer



from dataclasses import dataclass, field
from enum import Enum
from typing import Any, Type

import draccus
from draccus import ParsingError
from draccus.utils import DecodingError

from .testutils import TestSetup, parametrize, raises


def test_basic_required_argument(simple_attribute, silent):
    some_type, passed_value, expected_value = simple_attribute

    @dataclass
    class SomeDataclass(TestSetup):
        some_attribute: some_type  # type: ignore

    actual = SomeDataclass.setup(f"--some_attribute {passed_value}")
    assert actual.some_attribute == expected_value
    assert isinstance(actual.some_attribute, some_type)


def test_not_passing_required_argument_raises_error(simple_attribute):
    some_type, passed_value, expected_value = simple_attribute

    @dataclass
    class SomeDataclass(TestSetup):
        some_attribute: some_type  # type: ignore

    with raises(DecodingError):
        _ = SomeDataclass.setup("")


def test_basic_optional_argument(simple_attribute, silent):
    some_type, _, expected_value = simple_attribute

    @dataclass
    class SomeDataclass(TestSetup):
        some_attribute: some_type = expected_value  # type: ignore

    actual = SomeDataclass.setup("")
    assert actual.some_attribute == expected_value
    assert isinstance(actual.some_attribute, some_type)


@parametrize(
    "some_type, default_value,  arg_value",
    [
        (int, 0, 1234),
        (float, 0.0, 123.456),
        (str, "", "bobby_boots"),
        (bool, False, True),
    ],
)
def test_arg_value_is_set_when_args_are_provided(some_type: Type, default_value: Any, arg_value: Any):
    @dataclass
    class SomeClass(TestSetup):
        a: some_type = default_value  # type: ignore
        """some docstring for attribute 'a'"""

    class_a = SomeClass.setup(f"--a {arg_value}")
    assert class_a.a != default_value
    assert class_a.a == arg_value
    assert isinstance(class_a.a, some_type)


@parametrize(
    "some_type",
    [
        int,
        float,
        str,
        bool,
    ],
)
def test_not_providing_required_argument_throws_error(some_type):
    @dataclass
    class SomeClass(TestSetup):
        a: some_type  # type: ignore
        """some docstring for attribute 'a'"""

    with raises(ParsingError):
        _ = SomeClass.setup("")


@parametrize("some_type", [int, float, str])
def test_not_providing_required_argument_name_but_no_value_throws_error(some_type):
    @dataclass
    class SomeClass(TestSetup):
        a: some_type  # type: ignore
        """some docstring for attribute 'a'"""

    with raises(SystemExit):
        _ = SomeClass.setup("--a")


class Color(Enum):
    RED = "RED"
    ORANGE = "ORANGE"
    BLUE = "BLUE"


@dataclass
class Base(TestSetup):
    """A simple base-class example"""

    a: int  # TODO: finetune this
    """docstring for attribute 'a'"""
    b: float = 5.0  # inline comment on attribute 'b'
    c: str = ""


@dataclass
class Extended(Base):
    """Some extension of base-class `Base`"""

    d: int = 5
    """ docstring for 'd' in Extended. """
    e: Color = Color.BLUE


def test_parse_base_simple_works():
    b = Base.setup("--a 10 --b 3 --c Hello")
    assert b.a == 10
    assert b.b == 3
    assert b.c == "Hello"


def test_help_displays_class_docstring_text():
    assert Base.__doc__ in Base.get_help_text()


def test_enum_attributes_work():
    ext = Extended.setup("--a 5 --e RED")
    assert ext.e == Color.RED

    ext = Extended.setup("--a 5")
    assert ext.e == Color.BLUE


def test_parsing_twice():
    @dataclass
    class Foo:
        a: int = 123

    cfg = draccus.parse(config_class=Foo, args="")
    assert cfg.a == 123
    cfg = draccus.parse(config_class=Foo, args="--a 456".split())
    assert cfg.a == 456


def test_using_a_Type_type():
    @dataclass
    class Base:
        a: str = "a"

    @dataclass
    class Extended(Base):
        a: str = "extended_a"

    @dataclass
    class Foo(TestSetup):
        a_class: Type[Base] = field(default=Base, init=False)
        a: Base = field(default=None, init=False)

        def __post_init__(self):
            self.a = self.a_class()

    foo = Foo.setup("")
    from draccus.utils import contains_dataclass_type_arg

    assert not contains_dataclass_type_arg(Type[Base])
    assert foo.a_class() == Base()

    @dataclass
    class OtherFoo(Foo):
        a_class: Type[Base] = field(default=Extended, init=False)

    foo = OtherFoo.setup("")
    assert foo.a == Extended()


@dataclass(frozen=True)
class SomeClass:
    val: str


def test_argparse_with_custom_parsing():
    @dataclass
    class Config(TestSetup):
        s: SomeClass = SomeClass("bob")

    draccus.encode.register(SomeClass, lambda x: f"custom_{x.val}")
    draccus.decode.register(SomeClass, lambda x: SomeClass(x[7:]))

    Config.setup("--s custom_bob")


def test_empty_required_str():
    @dataclass
    class Args:
        a: str

    cfg = draccus.parse(Args, args=["--a="])

    assert cfg.a == ""


def test_exit_on_error():
    @dataclass
    class Args:
        a: int

    # Note: the cases that actually trigger the SystemExit are unclear.
    # Not all errors do. This could probably be improved.
    with raises(SystemExit):
        draccus.parse(Args, args=["--b=b"], exit_on_error=True)

    with raises(draccus.utils.DraccusException):
        draccus.parse(Args, args=["--b=b"], exit_on_error=False)



from dataclasses import dataclass, field

from .testutils import *


@dataclass(frozen=True)
class Base(TestSetup):
    """Some extension of base-class `Base`"""

    a: int = 1


@dataclass(frozen=True)
class ExtendedB(Base, TestSetup):
    b: int = 2


@dataclass(frozen=True)
class ExtendedC(Base, TestSetup):
    c: int = 3


@dataclass(frozen=True)
class Inheritance(TestSetup):
    ext_b: ExtendedB = ExtendedB()
    ext_c: ExtendedC = ExtendedC()


def test_simple_subclassing_no_args():
    extended = ExtendedB.setup()
    assert extended.a == 1
    assert extended.b == 2


def test_simple_subclassing_with_args():
    extended = ExtendedB.setup("--a 123 --b 56")
    assert extended.a == 123
    assert extended.b == 56


def test_subclasses_with_same_base_class_no_args():
    ext = Inheritance.setup()

    assert ext.ext_b.a == 1
    assert ext.ext_b.b == 2

    assert ext.ext_c.a == 1
    assert ext.ext_c.c == 3


def test_subclasses_with_same_base_class_with_args():
    import logging

    logging.basicConfig(level=logging.DEBUG)
    ext = Inheritance.setup("--ext_b.a 10 --ext_c.a 30")
    print(ext.get_help_text())
    assert ext.ext_b.a == 10
    assert ext.ext_c.a == 30


def test_weird_structure():
    """both is-a, and has-a at the same time, a very weird inheritance structure"""

    @dataclass
    class ConvBlock:
        """A Block of Conv Layers."""

        n_layers: int = 4  # number of layers
        n_filters: List[int] = field(default_factory=[16, 32, 64, 64].copy)  # filters per layer

    from enum import Enum

    class Optimizers(Enum):
        ADAM = "ADAM"
        RMSPROP = "RMSPROP"
        SGD = "SGD"

    @dataclass
    class GeneratorHParams(ConvBlock):
        """Settings of the Generator model"""

        conv: ConvBlock = field(default_factory=ConvBlock)
        optimizer: Optimizers = field(default=Optimizers.ADAM)

    @dataclass
    class DiscriminatorHParams(ConvBlock):
        """Settings of the Discriminator model"""

        conv: ConvBlock = field(default_factory=ConvBlock)
        optimizer: Optimizers = field(default=Optimizers.ADAM)

    @dataclass
    class SomeWeirdClass(TestSetup):
        gen: GeneratorHParams = field(default_factory=GeneratorHParams)
        disc: DiscriminatorHParams = field(default_factory=DiscriminatorHParams)

    s = SomeWeirdClass.setup()
    assert s.gen.conv.n_layers == 4
    assert s.gen.n_layers == 4
    assert s.disc.conv.n_layers == 4
    assert s.disc.n_layers == 4



from __future__ import annotations

from dataclasses import dataclass, field
from typing import List

import draccus

# it seems like typing.gettypehints doesn't really work with locals so, we just make these module scope


@dataclass(frozen=True)
class A:
    b: int = 1


@dataclass(frozen=True)
class C:
    a: A = A()
    elems: List[A] = field(default_factory=list)


def test_future_annotations():
    an_a: A = draccus.parse(config_class=A, args="")
    assert an_a.b == 1


def test_nested_future_annotations():
    c: C = draccus.parse(config_class=C, args="")
    assert c.a.b == 1



import argparse
import dataclasses
import inspect
from dataclasses import Field
from functools import cached_property
from typing import Dict, Optional, Sequence, Type

from ..choice_types import CHOICE_TYPE_KEY, ChoiceType
from ..parsers.decoding import has_custom_decoder
from ..utils import canonicalize_union, is_choice_type, is_union
from . import FieldWrapper, docstring
from .wrapper import AggregateWrapper, Wrapper


class ChoiceWrapper(AggregateWrapper[Type[ChoiceType]]):
    def __init__(
        self,
        choice_type: Type[ChoiceType],
        name: Optional[str] = None,
        # default: Optional[Union[Dataclass, Dict]] = None,
        parent: Optional[Wrapper] = None,
        _field: Optional[dataclasses.Field] = None,
        preferred_help: str = docstring.HelpOrder.inline,
    ):
        self.choice_type = choice_type
        self._name = name
        # TODO: need to use the default
        # self.default = default
        self._parent = parent
        self._field = _field

        # preferred parse for docstring / help text in < inline | above | below >
        self.preferred_help = preferred_help

        self._required: bool = False
        self._explicit: bool = False

    @property
    def title(self) -> str:
        title = self.choice_type.__qualname__
        if self.dest is not None:  # Show name if exists
            title += f" ['{self.dest}']"
        return title

    @property
    def parent(self) -> Optional["Wrapper"]:
        return self._parent

    @property
    def description(self) -> str:
        if self.parent and self._field:
            doc = docstring.get_attribute_docstring(self.parent.type, self._field.name)
            help_text = docstring.get_preferred_help_text(doc, preferred_help=self.preferred_help)
            if help_text is not None:
                return help_text
        class_doc = self.choice_type.__doc__ or ""
        if class_doc.startswith(f"{self.choice_type.__name__}("):
            return ""  # The base dataclass doc looks confusing, remove it
        return class_doc

    def register_actions(self, parser: argparse.ArgumentParser) -> None:
        group = parser.add_argument_group(title=self.title, description=self.description)
        children = self._children

        # Register the type argument. If closed, it's a choice between the known types, otherwise a string description
        dest = self.dest
        if dest is None:
            arg_name = CHOICE_TYPE_KEY
        else:
            arg_name = f"{dest}.{CHOICE_TYPE_KEY}"

        group.add_argument(
            f"--{arg_name}",
            choices=list(children.keys()),
            help=f"Which type of {self.title} to use",
            required=self.required,
        )

        for child in self._children.values():
            from .dataclass_wrapper import DataclassWrapper

            assert isinstance(child, DataclassWrapper)
            child.register_actions(parser)

    @cached_property
    def _children(self) -> Dict[str, Wrapper]:
        from .dataclass_wrapper import DataclassWrapper

        def _wrap_child(child: Type) -> Wrapper:
            if not dataclasses.is_dataclass(child):
                raise ValueError(f"Expected a dataclass, got {child}")
            if has_custom_decoder(child):
                raise ValueError(f"Cannot use class with custom decoder as choice type: {child}")

            # because we "substitute" the choice type for the child, we need to make sure that
            # the child's parent is the same as the choice type's parent
            return DataclassWrapper(child, parent=self.parent, _field=self._field, name=self.name)

        return {name: _wrap_child(child) for name, child in self.choice_type.get_known_choices().items()}

    @property
    def required(self) -> bool:
        return self._required

    @required.setter
    def required(self, value: bool):
        self._required = value
        for child_wrapper in self._children.values():
            child_wrapper.required = value

    @property
    def name(self) -> str:
        return self._name or self.choice_type.__name__

    @property
    def field(self) -> Optional[Field]:
        return self._field

    @property
    def type(self) -> Type:
        return self.choice_type


class UnionWrapper(AggregateWrapper[type]):
    def __init__(
        self,
        union: type,
        name: Optional[str] = None,
        parent: Optional[Wrapper] = None,
        _field: Optional[dataclasses.Field] = None,
        preferred_help: str = docstring.HelpOrder.inline,
    ):
        self.union = canonicalize_union(union)
        self._name = name
        self._parent = parent
        self._field = _field

        # preferred parse for docstring / help text in < inline | above | below >
        self.preferred_help = preferred_help

        self._required: bool = False

        # if None is in the union, then it's optional
        self._required = type(None) not in self.union.__args__

        self._explicit: bool = False

    @property
    def name(self) -> str:
        return self._name or self.union.__name__

    @property  # type: ignore
    def required(self) -> bool:  # type: ignore
        return self._required

    @property
    def field(self) -> Optional[Field]:
        return self._field

    @property
    def type(self) -> Type:
        return self.union

    @property
    def title(self) -> str:
        try:
            title = self.union.__qualname__
        except AttributeError:
            try:
                title = self.union.__name__
            except AttributeError:
                title = str(self.union)
        if self.dest is not None:  # Show name if exists
            title += f" ['{self.dest}']"
        return title

    @property
    def parent(self) -> Optional["Wrapper"]:
        return self._parent

    @property
    def description(self) -> str:
        if self.parent and self._field:
            doc = docstring.get_attribute_docstring(self.parent.type, self._field.name)
            help_text = docstring.get_preferred_help_text(doc, preferred_help=self.preferred_help)
            if help_text is not None:
                return help_text
        class_doc = self.union.__doc__ or ""
        return class_doc

    def register_actions(self, parser: argparse.ArgumentParser) -> None:
        # In Pyrallis/Draccus, Unions are implicitly resolved with no tag, unlike Choices
        group = parser.add_argument_group(title=self.title, description=self.description)
        children = self._children

        has_field_wrapper = False

        for child in children:
            from .dataclass_wrapper import DataclassWrapper

            if isinstance(child, DataclassWrapper):
                child.register_actions(parser)
            elif isinstance(child, FieldWrapper):
                has_field_wrapper = True
            elif child is None:
                pass
            elif isinstance(child, ChoiceWrapper):
                child.register_actions(parser)
            else:
                raise ValueError(f"Unexpected child type: {child}")

        if has_field_wrapper:
            if self._field is None:
                help_text: Optional[str] = None
            elif self.parent is not None and self._field is not None:
                doc = docstring.get_attribute_docstring(self.parent.type, self._field.name)
                help_text = docstring.get_preferred_help_text(doc, preferred_help=self.preferred_help)
            else:
                help_text = None

            group.add_argument(
                f"--{self.dest}",
                required=False,
                help=help_text,
            )

    @cached_property
    def _children(self) -> Sequence[Optional[Wrapper]]:
        from .dataclass_wrapper import DataclassWrapper
        from .field_wrapper import FieldWrapper

        def _wrap_child(child: Type) -> Optional[Wrapper]:
            if has_custom_decoder(child):
                assert self._field is not None
                field = FieldWrapper(
                    parent=self.parent, field=self._field, preferred_help=self.preferred_help, field_type=child
                )
                field.required = False
                return field
            elif is_choice_type(child):
                return ChoiceWrapper(
                    child, name=self.name, parent=self.parent, _field=self._field, preferred_help=self.preferred_help
                )
            elif dataclasses.is_dataclass(child):
                return DataclassWrapper(
                    child, name=self.name, parent=self.parent, _field=self._field, preferred_help=self.preferred_help
                )
            elif is_union(child):
                return UnionWrapper(child, parent=self.parent, _field=self._field, preferred_help=self.preferred_help)
            elif child is None or child is type(None):
                return None
            else:
                assert self._field is not None
                wrapper = FieldWrapper(
                    parent=self.parent, field=self._field, preferred_help=self.preferred_help, field_type=child
                )
                wrapper.required = False
                return wrapper

        return [_wrap_child(child) for child in self.union.__args__]
